"""Top-level package for chat-replay-downloader."""

__author__ = 'Joshua Lochner'
__email__ = 'admin@xenova.com'
__version__ = '0.0.9'
__url__ = 'https://github.com/xenova/chat-replay-downloader'

from .cli import main
from .chat_replay_downloader import ChatDownloader
